%% Setup

% Dependent functions
addpath('utilities');
addpath('../Tools');

%% Settings

formationThreshold = 0.1; % Error threshold for when defined as "in formation"

% Dynamics
dyn_simple = 0; % Use single integrator kinematics
m = 3; % Mass

% Controller
cont_pos = 1;   % Use position-based formation control
    cont_coll_av = 0;
        r_0 = 3; % Repulsion radius
        r_1 = 2.5; % Repulsion saturation
        K_coll = 20*m; % Repulsion gain
K_miss = 2*m; % Mission velocity gain (damping on u)
delta = 0.5*m; % Link gain
V_max = 3; % Maximum velocity
F_max = 10*m; % Maximum force

% Guidance
%v_const = [-0.2 -0.2 0]';
%v_const = [0 0 0]';

% Time
t_end = 6;
h = 0.01;
time = 0:h:t_end;

% Plotting
plotView = [-120,30]; % [Az, El]
%plotView = [109,54];
marg = 1;
anim = 0;
    movie = 0;
        name = 'movies\test';


%% Initial conditions, communication topology and desired formation

% Get init pos, incidence matrix and desired pos
%form_triangle;
%form_V;
%form_example; % Example for report

[N,L] = size(D); % [Agents, Links]
R = size(x_0,1); % Spatial dimentions

% Desired difference variables [link vectors]
z_d = kron(D',eye(R))*x_c(:);
z_d = reshape(z_d,R,L);

% Desired link lengths
d_l = sqrt(sum(z_d.^2,1));

% Link gains
deltas = delta*ones(L,1);


%% Run simulation loop


% Initialize Logger
logger = Logger();
logger.init(time,h);
logger.add('x', R*N);
logger.add('v', R*N);
logger.add('u', R*N);
logger.add('z', R*L);
logger.add('tau', R*N);
logger.add('v_mission', R);
logger.add('a_pot', R*N);
logger.add('a_form', R*N);
logger.add('inFormation', N);

% Initialize variables
x = x_0;
x_next = zeros(R,N);
v = zeros(R,N);
v_next = zeros(R,N);
tau = zeros(R,N);
a_pot = zeros(R,N);

a_max = zeros(R,1);
d_ij_min = inf(1,N);

% Simulation loop
for k = 1:length(time)
    %time(k)
    
    z = kron(D',eye(R))*x(:);
    z = reshape(z,R,L);
    
    if cont_pos
        
        % Position-based formation control
        z_tilde = z - z_d;
        psi = z_tilde(:).*kron(deltas,ones(R,1));
        a_form = -kron(D,eye(R))*psi(:);
        a_form = reshape(a_form,R,N);
        
        if cont_pos && cont_coll_av
            
            % Repulsion from collision avoidance
            a_pot = zeros(R,N);
            for agent = 1:N
                x_i = x(:,agent);
                for neighbour = 1:N
                    if agent ~= neighbour
                        x_j = x(:,neighbour);
                        x_ij = x_j - x_i;
                        d_ij = norm(x_ij);
                        if d_ij < r_0
                            a_pot(:,agent) = a_pot(:,agent) - min(r_1, r_0 - d_ij)*x_ij/d_ij;
                        end
                        d_ij_min(agent) = min(d_ij_min(agent), d_ij);
                    end
                end
                a_pot(:,agent) = K_coll*a_pot(:,agent);
                a_max = max(a_max, a_pot(:,agent));
            end
        end
        u = a_form + a_pot;
    else
        
        % Distance-based formation control
        psi = zeros(R,L);
        for link = 1:L
            z_l_abs = norm(z(:,link));
            %psi(:,link) = deltas(link)*z(:,link)*(z_l_abs - d_l(link))/z_l_abs^2;
            psi(:,link) = deltas(link)*z(:,link)*log(z_l_abs/d_l(link))/z_l_abs;
        end
        a_form = -kron(D,eye(R))*psi(:);
        a_form = reshape(a_form,R,N);
        u = a_form;
    end
    
    
    % In formation?
    inFormation = 1;
    for link = 1:L
        if norm(z_tilde(:,link)) > formationThreshold
            inFormation = 0;
            break
        end
    end  
    
    
    % Mission velocity
    v_mission = v_const;
    dv_mission = [0 0 0]';
    
    
    for agent = 1:N
        x_i = x(:,agent);
        v_i = v(:,agent);
        u_i = u(:,agent);
        
        if dyn_simple
            tau_i = u_i + v_mission;
            tau_i = saturate(tau_i, V_max); % Limit speed
            
            f = @(x,v)(v);
            x_i_next = rk4(f, x_i, v_i, h);
            
            v_i_next = tau_i;
        else
            tau_i = -K_miss*(v_i - v_mission) + m*dv_mission + u_i;
            tau_i = saturate(tau_i, F_max); % Limit force
            
            f = @(x,v)(v);
            x_i_next = rk4(f, x_i, v_i, h);           
            
            f = @(v,tau)(tau/m);
            %v_i_next = rk4(f, v_i, tau_i, h);
            v_i_next = saturate(rk4(f, v_i, tau_i, h),V_max); % Limit speed (valid?)
        end
        
        x_next(:,agent) = x_i_next;
        v_next(:,agent) = v_i_next;
        tau(:,agent) = tau_i;
    end
    x = x_next;
    v = v_next;
    
    % Store variables
    logger.store('x', x(:), k);
    logger.store('v', v(:), k);
    logger.store('u', u(:), k);
    logger.store('z', z(:), k);
    logger.store('tau', tau(:), k);
    logger.store('v_mission', v_mission, k);
    logger.store('a_pot', a_pot(:), k);
    logger.store('a_form', a_form(:), k);
    logger.store('inFormation', inFormation, k);
end
        
        

%% Plots

% Find plot limits
all_pos = reshape(logger.get('x'),R,[]);
lim_x = [min(all_pos(1,:))-marg, max(all_pos(1,:))+marg];
lim_y = [min(all_pos(2,:))-marg, max(all_pos(2,:))+marg];
lim_z = [min(all_pos(3,:))-marg, max(all_pos(3,:))+marg];

coordForEach = @(i)(R*(i-1)+1:R*(i-1)+R);

colors = {'b','r','k','g','m','c'};
col = get(groot,'DefaultAxesColorOrder');
dims = {'N', 'E', 'D'};
set(0,'defaulttextinterpreter','tex')


% % 3D and summary plot
%plotFormation;


% Animated 3D plot
if anim
    animateFormation;
end

    
    
   